@@ -47,9 +47,23 @@
     Updated by      Richard Gooch   13-JUN-1999: Added compile-time check of
   protocol revision.
 
-    Last updated by Richard Gooch   14-JUN-1999: Stat inode instead of using
+    Updated by      Richard Gooch   14-JUN-1999: Stat inode instead of using
   device entry information (which may be out of date).
 
+    Updated by      Richard Gooch   15-JUN-1999: Added tracing output. Thanks
+  to Piete Brooks <Piete.Brooks@cl.cam.ac.uk>.
+
+    Updated by      Richard Gooch   16-JUN-1999: Added action structure.
+
+    Updated by      Richard Gooch   17-JUN-1999: Added "hostname" and "mntpnt"
+  variables. Added nesting of include files. Added "CLEAR_CONFIG" key.
+
+    Updated by      Richard Gooch   18-JUN-1999: Added "OPTIONAL_INCLUDE" key.
+  Added ignore actions. Added event mask.
+
+    Last updated by Richard Gooch   24-JUN-1999: Switched to separated event
+  and action config file specifiers.
+
 
 */
 #include <stdio.h>
@@ -74,12 +88,12 @@
 #include <karma.h>
 
 
-#define VERSION "1.2.2"
+#define VERSION "1.2.3"
 #define CONFIG_FILE "/etc/devfsd.conf"
 #define MAX_ARGS (6 + 1)
 
 /*  Update only after changing code to reflect new protocol  */
-#define DEVFSD_PROTOCOL_REVISION_DAEMON  3
+#define DEVFSD_PROTOCOL_REVISION_DAEMON  4
 
 /*  Compile-time check  */
 #if DEVFSD_PROTOCOL_REVISION_KERNEL != DEVFSD_PROTOCOL_REVISION_DAEMON
@@ -87,15 +101,15 @@
 #endif
 
 
-#define ACTION_PERMISSIONS              0x0000
-#define ACTION_EXECUTE_ON_REGISTER      0x0100
-#define ACTION_EXECUTE_ON_UNREGISTER    0x0200
-#define ACTION_EXECUTE_ON_ASYNC_OPEN    0x0300
-#define ACTION_EXECUTE_ON_CLOSE         0x0400
-#define ACTION_EXECUTE_MASK             0xff00
+#define AC_PERMISSIONS              0
+#define AC_EXECUTE                  1
+#define AC_IGNORE                   2
+
+#define MKACTION(what,when) (struct action_type) {what, when}
 
 #define SYSLOG(level, format, args...) \
-    {if (trace_level > 0) fprintf (stderr, format, ## args); \
+    {if ( debug_protocol || (trace_level > 0) ) \
+         fprintf (stderr, format, ## args); \
      else syslog (level, format, ##args); \
     }
 
@@ -112,9 +126,15 @@ struct execute_type
     char *argv[MAX_ARGS + 1];
 };
 
+struct action_type
+{
+    unsigned int what;
+    unsigned int when;
+};
+
 struct config_entry
 {
-    unsigned int action;
+    struct action_type action;
     regex_t preg;
     union
     {
@@ -125,6 +145,12 @@ struct config_entry
     struct config_entry *next;
 };
 
+struct get_variable_info
+{
+    const struct devfsd_notify_struct *info;
+    CONST char *devname;
+    char devpath[STRING_LENGTH];
+};
 
 
 /*  External functions  */
@@ -137,8 +163,9 @@ EXTERN_FUNCTION (flag st_expr_expand,
 
 /*  Private functions  */
 static void setup_initial_entries (CONST char *mount_point);
-static void read_config ();
-static void do_servicing (int fd);
+static void read_config (CONST char *filename, flag optional,
+			 unsigned long *event_mask);
+static void do_servicing (int fd, unsigned long event_mask);
 static void service_name (const struct devfsd_notify_struct *info);
 static void update_permissions (const struct devfsd_notify_struct *info,
 				const struct config_entry *entry);
@@ -158,6 +185,7 @@ static struct config_entry *first_config = NULL;
 static struct config_entry *last_config = NULL;
 static const char *mount_point = NULL;
 static char trace_level = 0;
+static flag debug_protocol = FALSE;
 static struct initial_symlink_struct
 {
     char *dest;
@@ -177,7 +205,6 @@ static struct initial_symlink_struct
 int main (int argc, char **argv)
 {
     flag print_version = FALSE;
-    flag debug = FALSE;
     flag debug_devfs = FALSE;
     int fd, proto_rev, count;
     unsigned int devfs_debug_mask;
@@ -194,10 +221,10 @@ int main (int argc, char **argv)
     for (count = 2; count < argc; ++count)
     {
 	if (strcmp (argv[count], "-v") == 0) print_version = TRUE;
-	else if (strcmp (argv[count], "-d") == 0) debug = TRUE;
-	else if ( (strcmp (argv[count], "-t") == 0) && (++count < argc) )
+	else if (strcmp (argv[count], "-d") == 0) debug_protocol = TRUE;
+	else if ( (strcmp (argv[count], "-t") == 0) && (count + 1 < argc) )
 	{
-	    trace_level = atoi (argv[count]);
+	    trace_level = atoi (argv[++count]);
 	}
 	else if ( (strcmp (argv[count], "-D") == 0) && (++count < argc) )
 	{
@@ -224,7 +251,24 @@ int main (int argc, char **argv)
 	fprintf (stderr, "Error getting protocol revision\t%s\n", ERRSTRING);
 	exit (1);
     }
-    if ( print_version || (trace_level > 0) || debug ||
+    setup_initial_entries (mount_point);
+    if (!print_version && !debug_protocol && !debug_devfs)
+    {
+	if (stat (CONFIG_FILE, &statbuf) != 0)
+	{
+	    fprintf (stderr, "device management daemon exiting as no %s\n",
+		     CONFIG_FILE);
+	    exit (0);
+	}
+	if (statbuf.st_size == 0)
+	{
+	    fprintf (stderr,
+		     "device management daemon exiting as %s is empty\n",
+		     CONFIG_FILE);
+	    exit (0);
+	}
+    }
+    if ( print_version || (trace_level > 0) || debug_protocol ||
 	 (DEVFSD_PROTOCOL_REVISION_DAEMON != proto_rev) )
     {
 	fprintf (stderr, "devfsd: Linux device filesystem daemon v%s\n",
@@ -250,8 +294,7 @@ int main (int argc, char **argv)
 	}
 	exit (0);
     }
-    if (debug) do_debug (fd);
-    setup_initial_entries (mount_point);
+    if (debug_protocol) do_debug (fd);
     sigemptyset (&new_action.sa_mask);
     new_action.sa_flags = 0;
     /*  Set up SIGHUP handler  */
@@ -261,11 +304,10 @@ int main (int argc, char **argv)
 	fprintf (stderr, "Error setting SIGHUP handler\t%s\n", ERRSTRING);
 	exit (1);
     }
-    if (stat (CONFIG_FILE, &statbuf) != 0) exit (0);
-    if (statbuf.st_size == 0) exit (0);
-    fprintf (stderr, "Started device management daemon for %s\n", mount_point);
     if (trace_level < 1)
     {
+    	fprintf (stderr, "Started device management daemon for %s\n",
+		 mount_point);
 	switch ( fork () )
 	{
 	  case 0:
@@ -286,11 +328,19 @@ int main (int argc, char **argv)
 	close (1);
 	close (2);
     }
+    else
+    {
+    	fprintf (stderr,
+		 "Started device management daemon for %s at trace level %d\n",
+		 mount_point, trace_level);
+    }
     while (TRUE)
     {
-	read_config ();
+	unsigned long event_mask = 0;
+
+	read_config (CONFIG_FILE, FALSE, &event_mask);
 	/*  Carry on daemon  */
-	do_servicing (fd);
+	do_servicing (fd, event_mask);
 	free_config ();
     }
 }   /*  End Function main  */
@@ -312,8 +362,12 @@ static void setup_initial_entries (CONST char *mount_point)
     }
 }   /*  End Function setup_initial_entries  */
 
-static void read_config ()
+static void read_config (CONST char *filename, flag optional,
+			 unsigned long *event_mask)
 /*  [SUMMARY] Read the configuration file.
+    <filename> The filename to read from.
+    <optional> If TRUE, the routine will silently ignore a missing config file.
+    <event_mask> The event mask is written here. This is not initialised.
     [RETURNS] Nothing.
 */
 {
@@ -321,17 +375,21 @@ static void read_config ()
     FILE *fp;
     char buf[STRING_LENGTH];
 
-    if ( ( fp = fopen (CONFIG_FILE, "r") ) == NULL )
+    if ( ( fp = fopen (filename, "r") ) == NULL )
     {
+	if ( optional && (errno == ENOENT) ) return;
 	SYSLOG (LOG_ERR, "error opening file: \"%s\"\t%s\n",
 		CONFIG_FILE, ERRSTRING);
 	exit (1);
     }
+    if (trace_level > 1) fprintf (stderr, "Reading file: \"%s\"\n",
+				  CONFIG_FILE);
     while (fgets (buf, STRING_LENGTH, fp) != NULL)
     {
 	int err, num_args, count;
 	struct config_entry *new;
-	char p[MAX_ARGS][STRING_LENGTH], key[STRING_LENGTH];
+	char p[MAX_ARGS][STRING_LENGTH];
+	char when[STRING_LENGTH], what[STRING_LENGTH];
 	char name[STRING_LENGTH], tmp[STRING_LENGTH];
 
 	buf[strlen (buf) - 1] = '\0';
@@ -340,15 +398,36 @@ static void read_config ()
 	if (line[0] == '\0') continue;
 	if (line[0] == '#') continue;
 	for (count = 0; count < MAX_ARGS; ++count) p[count][0] = '\0';
-	num_args = sscanf (line, "%s %s %s %s %s %s %s %s %s",
-			   key, name, p[0], p[1], p[2], p[3],
+	num_args = sscanf (line, "%s %s %s %s %s %s %s %s %s %s",
+			   when, what, name, p[0], p[1], p[2], p[3],
 			   p[4], p[5], p[6]);
+	if (strcasecmp (when, "CLEAR_CONFIG") == 0)
+	{
+	    free_config ();
+	    *event_mask = 0;
+	    continue;
+	}
 	if (num_args < 2)
 	{
 	    SYSLOG (LOG_ERR, "bad config line: \"%s\"\n", line);
 	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
+	if ( (strcasecmp (when, "INCLUDE") == 0) ||
+	     (strcasecmp (when, "OPTIONAL_INCLUDE") == 0) )
+	{
+	    st_expr_expand (name, STRING_LENGTH, what, get_variable, NULL,
+			    NULL);
+	    read_config (what, (toupper (when[0]) == 'I') ? FALSE : TRUE,
+			 event_mask);
+	    continue;
+	}
+	if (num_args < 3)
+	{
+	    SYSLOG (LOG_ERR, "bad config line: \"%s\"\n", line);
+	    SYSLOG (LOG_ERR, "exiting\n");
+	    exit (1);
+	}
 	if ( ( new = malloc (sizeof *new) ) == NULL )
 	{
 	    SYSLOG (LOG_ERR, "error allocating\n");
@@ -356,32 +435,29 @@ static void read_config ()
 	    exit (1);
 	}
 	memset (new, 0, sizeof *new);
-	if (strcasecmp (key, "PERMISSIONS") == 0)
-	    new->action = ACTION_PERMISSIONS;
-	else if (strcasecmp (key, "REG_EXE") == 0)
-	    new->action = ACTION_EXECUTE_ON_REGISTER;
-	else if (strcasecmp (key, "UNREG_EXE") == 0)
-	    new->action = ACTION_EXECUTE_ON_UNREGISTER;
-	else if (strcasecmp (key, "ASYNC_OPEN_EXE") == 0)
-	    new->action = ACTION_EXECUTE_ON_ASYNC_OPEN;
-	else if (strcasecmp (key, "CLOSE_EXE") == 0)
-	    new->action = ACTION_EXECUTE_ON_CLOSE;
+	if (strcasecmp (when, "REGISTER") == 0)
+	    new->action.when = DEVFSD_NOTIFY_REGISTERED;
+	else if (strcasecmp (when, "UNREGISTER") == 0)
+	    new->action.when = DEVFSD_NOTIFY_UNREGISTERED;
+	else if (strcasecmp (when, "ASYNC_OPEN") == 0)
+	    new->action.when = DEVFSD_NOTIFY_ASYNC_OPEN;
+	else if (strcasecmp (when, "CLOSE") == 0)
+	    new->action.when = DEVFSD_NOTIFY_CLOSE;
+	else if (strcasecmp (when, "LOOKUP") == 0)
+	    new->action.when = DEVFSD_NOTIFY_LOOKUP;
+	else if (strcasecmp (when, "CHANGE") == 0)
+	    new->action.when = DEVFSD_NOTIFY_CHANGE;
+	else if (strcasecmp (when, "CREATE") == 0)
+	    new->action.when = DEVFSD_NOTIFY_CREATE;
 	else
 	{
-	    SYSLOG (LOG_ERR, "bad config line: \"%s\"\n", line);
+	    SYSLOG (LOG_ERR, "bad WHEN in config line: \"%s\"\n", line);
 	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
-	if ( ( err = regcomp (&new->preg, name, 0) ) != 0 )
+	if (strcasecmp (what, "PERMISSIONS") == 0)
 	{
-	    regerror (err, &new->preg, tmp, STRING_LENGTH);
-	    SYSLOG (LOG_ERR, "error compiling regexp: \"%s\"\t%s\n", name,tmp);
-	    SYSLOG (LOG_ERR, "exiting\n");
-	    exit (1);
-	}
-	switch (new->action)
-	{
-	  case ACTION_PERMISSIONS:
+	    new->action.what = AC_PERMISSIONS;
 	    /*  Get user and group  */
 	    if ( ( ptr = strchr (p[0], '.') ) == NULL )
 	    {
@@ -394,12 +470,11 @@ static void read_config ()
 	    new->u.permissions.gid = get_gid (ptr);
 	    /*  Get mode  */
 	    new->u.permissions.mode = get_mode (p[1]);
-	    break;
-	  case ACTION_EXECUTE_ON_REGISTER:
-	  case ACTION_EXECUTE_ON_UNREGISTER:
-	  case ACTION_EXECUTE_ON_ASYNC_OPEN:
-	  case ACTION_EXECUTE_ON_CLOSE:
-	    num_args -= 2;
+	}
+	else if (strcasecmp (what, "EXECUTE") == 0)
+	{
+	    new->action.what = AC_EXECUTE;
+	    num_args -= 3;
 	    for (count = 0; count < num_args; ++count)
 	    {
 		err = strlen (p[count]) + 1;
@@ -412,8 +487,23 @@ static void read_config ()
 		strcpy (new->u.execute.argv[count], p[count]);
 	    }
 	    new->u.execute.argv[num_args] = NULL;
-	    break;
 	}
+	else if (strcasecmp (what, "IGNORE") == 0)
+	    new->action.what = AC_IGNORE;
+	else
+	{
+	    SYSLOG (LOG_ERR, "bad WHAT in config line: \"%s\"\n", line);
+	    SYSLOG (LOG_ERR, "exiting\n");
+	    exit (1);
+	}
+	if ( ( err = regcomp (&new->preg, name, 0) ) != 0 )
+	{
+	    regerror (err, &new->preg, tmp, STRING_LENGTH);
+	    SYSLOG (LOG_ERR, "error compiling regexp: \"%s\"\t%s\n", name,tmp);
+	    SYSLOG (LOG_ERR, "exiting\n");
+	    exit (1);
+	}
+	*event_mask |= 1 << new->action.when;
 	new->next = NULL;
 	if (first_config == NULL) first_config = new;
 	else last_config->next = new;
@@ -423,15 +513,22 @@ static void read_config ()
     SYSLOG (LOG_INFO, "read config file: \"%s\"\n", CONFIG_FILE);
 }   /*  End Function read_config   */
 
-static void do_servicing (int fd)
+static void do_servicing (int fd, unsigned long event_mask)
 /*  [SUMMARY] Service devfs changes until a signal is received.
     <fd> The open control file.
+    <event_mask> The event mask.
     [RETURNS] Nothing.
 */
 {
     ssize_t bytes;
     struct devfsd_notify_struct info;
 
+    /*  Tell devfs what events we care about  */
+    if (ioctl (fd, DEVFSDIOC_SET_EVENT_MASK, event_mask) != 0)
+    {
+	SYSLOG (LOG_ERR, "error setting event mask\t%s\n", ERRSTRING);
+	exit (1);
+    }
     while ( ( bytes = read (fd, (char *) &info, sizeof info) ) >= 0 )
     {
 	if (bytes < 1)
@@ -455,49 +552,31 @@ static void service_name (const struct devfsd_notify_struct *info)
     regmatch_t mbuf;
     struct config_entry *entry;
 
+    if (info->overrun_count > 0)
+	SYSLOG (LOG_ERR, "%u events have been lost!\n", info->overrun_count);
+    if (trace_level > 1) fprintf (stderr, "Looking for \"%s\" (%d)\n",
+				  info->devname, info->type);
     for (entry = first_config; entry != NULL; entry = entry->next)
     {
+	if (trace_level > 2) fprintf (stderr, "\t\tProcess \"%s\" (%d)\n",
+				      info->devname, entry->action.when);
+	/*  First check if action matches the type, then check if name matches
+	 */
+	if (info->type != entry->action.when) continue;
 	if (regexec (&entry->preg, info->devname, 1, &mbuf, 0) != 0) continue;
-	/*  We have a match, does the action match the type?  */
-	switch (info->type)
+	switch (entry->action.what)
 	{
-	  case DEVFSD_NOTIFY_REGISTERED:
-	    switch (entry->action)
-	    {
-	      case ACTION_PERMISSIONS:
-		update_permissions (info, entry);
-		break;
-	      case ACTION_EXECUTE_ON_REGISTER:
-		execute (info, entry);
-		break;
-	    }
+	  case AC_PERMISSIONS:
+	    update_permissions (info, entry);
 	    break;
-	  case DEVFSD_NOTIFY_UNREGISTERED:
-	    switch (entry->action)
-	    {
-	      case ACTION_EXECUTE_ON_UNREGISTER:
-		execute (info, entry);
-		break;
-	    }
-	    break;
-	  case DEVFSD_NOTIFY_ASYNC_OPEN:
-	    switch (entry->action)
-	    {
-	      case ACTION_EXECUTE_ON_ASYNC_OPEN:
-		execute (info, entry);
-		break;
-	    }
-	    break;
-	  case DEVFSD_NOTIFY_CLOSE:
-	    switch (entry->action)
-	    {
-	      case ACTION_EXECUTE_ON_CLOSE:
-		execute (info, entry);
-		break;
-	    }
+	  case AC_EXECUTE:
+	    execute (info, entry);
 	    break;
+	  case AC_IGNORE:
+	    return;
+	    /*break;*/
 	  default:
-	    SYSLOG (LOG_ERR, "Unknown event type: %u\n", info->type);
+	    SYSLOG (LOG_ERR, "Unknown action type: %u\n", entry->action.what);
 	    exit (1);
 	    /*break;*/
 	}
@@ -528,6 +607,11 @@ static void update_permissions (const struct devfsd_notify_struct *info,
     }
     new_mode = (statbuf.st_mode & S_IFMT) |
 	(entry->u.permissions.mode & ~S_IFMT);
+    if (trace_level > 1)
+	fprintf (stderr,
+		 "\tupdate permissions for \"%s\" from %05o to %05o, user.goup from %d.%d to %d.%d\n",
+		 path, info->mode, new_mode, info->uid, info->gid,
+		 entry->u.permissions.uid, entry->u.permissions.gid);
     if (new_mode != statbuf.st_mode)
     {
 	if (chmod (path, new_mode) != 0)
@@ -537,6 +621,8 @@ static void update_permissions (const struct devfsd_notify_struct *info,
 	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
+	if (trace_level > 1) fprintf (stderr,"Set permission %05o on \"%s\"\n",
+				      new_mode, path);
     }
     if ( (entry->u.permissions.uid != statbuf.st_uid) ||
 	 (entry->u.permissions.gid != statbuf.st_gid) )
@@ -549,6 +635,9 @@ static void update_permissions (const struct devfsd_notify_struct *info,
 	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
+	if (trace_level > 1)
+	    fprintf (stderr, "Set user.group %d.%d on \"%s\"\n",
+		     entry->u.permissions.uid, entry->u.permissions.gid, path);
     }
 }   /*  End Function update_permissions  */
 
@@ -561,20 +650,28 @@ static void execute (const struct devfsd_notify_struct *info,
 */
 {
     unsigned int count;
+    struct get_variable_info gv_info;
     char *argv[MAX_ARGS + 1];
     char largv[MAX_ARGS + 1][STRING_LENGTH];
-    char devname[STRING_LENGTH];
 
-    strcpy (devname, mount_point);
-    strcat (devname, "/");
-    strcat (devname, info->devname);
+    gv_info.info = info;
+    gv_info.devname = info->devname;
+    strcpy (gv_info.devpath, mount_point);
+    strcat (gv_info.devpath, "/");
+    strcat (gv_info.devpath, info->devname);
+    if (trace_level > 0) fprintf (stderr, "Calling %s for \"%s\" | \"%s\":",
+				  entry->u.execute.argv[0], info->devname,
+				  gv_info.devpath);
     for (count = 0; entry->u.execute.argv[count] != NULL; ++count)
     {
 	st_expr_expand (largv[count], STRING_LENGTH,
-			entry->u.execute.argv[count], get_variable, devname,
+			entry->u.execute.argv[count], get_variable, &gv_info,
 			NULL);
 	argv[count] = largv[count];
+	if (trace_level > 0) fprintf (stderr, " '%s'", argv[count]);
     }
+    if (trace_level > 0) fprintf (stderr, ". i.e. %d arg%s\n",
+				  count-1, (count == 2) ? "" : "s");
     argv[count] = NULL;
     switch ( fork () )
     {
@@ -609,7 +706,7 @@ static void free_config ()
 
 	next = entry->next;
 	regfree (&entry->preg);
-	if (entry->action & ACTION_EXECUTE_MASK)
+	if (entry->action.what == AC_EXECUTE)
 	{
 	    for (count = 0; count < MAX_ARGS; ++count)
 	    {
@@ -631,9 +728,15 @@ static void do_debug (int fd)
 {
     ssize_t bytes;
     struct devfsd_notify_struct info;
-    char type[STRING_LENGTH];
+    char *type;
 
     fprintf (stderr, "Running in debug mode: no actions will be taken...\n");
+    /*  Tell devfs what events we care about  */
+    if (ioctl (fd, DEVFSDIOC_SET_EVENT_MASK, ~0) != 0)
+    {
+	SYSLOG (LOG_ERR, "error setting event mask\t%s\n", ERRSTRING);
+	exit (1);
+    }
     while ( ( bytes = read (fd, (char *) &info, sizeof info) ) >= 0 )
     {
 	if (bytes < 1)
@@ -641,28 +744,39 @@ static void do_debug (int fd)
 	    fprintf (stderr, "devfs closed file!\n");
 	    exit (1);
 	}
+	if (info.overrun_count > 0)
+	    SYSLOG (LOG_ERR, "%u events have been lost!\n",info.overrun_count);
 	switch (info.type)
 	{
 	  case DEVFSD_NOTIFY_REGISTERED:
-	    strcpy (type, "registered");
+	    type = "registered";
 	    break;
 	  case DEVFSD_NOTIFY_UNREGISTERED:
-	    strcpy (type, "unregistered");
+	    type = "unregistered";
 	    break;
 	  case DEVFSD_NOTIFY_ASYNC_OPEN:
-	    strcpy (type, "asynchonously opened");
+	    type = "asynchonously opened";
 	    break;
 	  case DEVFSD_NOTIFY_CLOSE:
-	    strcpy (type, "closed");
+	    type = "closed";
+	    break;
+	  case DEVFSD_NOTIFY_LOOKUP:
+	    type = "lookup";
+	    break;
+	  case DEVFSD_NOTIFY_CHANGE:
+	    type = "change";
+	    break;
+	  case DEVFSD_NOTIFY_CREATE:
+	    type = "create";
 	    break;
 	  default:
 	    fprintf (stderr, "Unknown type: %u\n", info.type);
 	    exit (1);
 	    /*break;*/
 	}
-	fprintf (stderr, "Entry: \"%s\" %s mode: %o uid: %d gid: %d\n",
-		 info.devname, type, (int) info.mode, (int) info.uid,
-		 (int) info.gid);
+	fprintf (stderr, "Entry: \"%s\"(len=%d) %s mode: %o uid: %d gid: %d\n",
+		 info.devname, strlen (info.devname), type,
+		 (int) info.mode, (int) info.uid, (int) info.gid);
     }
     fprintf (stderr, "Error reading control file\t%s\n", ERRSTRING);
     exit (1);
@@ -729,6 +843,8 @@ static mode_t get_mode (const char *string)
     if (string[6] == 'r') mode |= S_IROTH;
     if (string[7] == 'w') mode |= S_IWOTH;
     if (string[8] == 'x') mode |= S_IXOTH;
+    if (trace_level > 2) fprintf (stderr, "\tMode \"%s\" gives %05o\n",
+				  string, mode);
     return (mode);
 }   /*  End Function get_mode  */
 
@@ -739,8 +855,39 @@ static void sig_handler (int sig)
 
 static CONST char *get_variable (CONST char *variable, void *info)
 {
-    if (strcmp (variable, "devpath") == 0) return (info);
-    if (strcmp (variable, "devname") == 0) return strchr (info, '/') + 1;
+    struct get_variable_info *gv_info = info;
+    static flag first_time = TRUE;
+    static char hostname[STRING_LENGTH], sbuf[STRING_LENGTH];
+
+    if (first_time)
+    {
+	first_time = FALSE;
+	if (gethostname (hostname, STRING_LENGTH - 1) != 0)
+	{
+	    SYSLOG (LOG_ERR, "Error getting hostname\t%s\n", ERRSTRING);
+	    exit (RV_SYS_ERROR);
+	}
+	hostname[STRING_LENGTH - 1] = '\0';
+    }
+    if (strcmp (variable, "hostname") == 0) return (hostname);
+    if (strcmp (variable, "mntpnt") == 0) return (mount_point);
+    if (gv_info == NULL) return (NULL);
+    if (strcmp (variable, "devpath") == 0) return (gv_info->devpath);
+    if (strcmp (variable, "devname") == 0) return (gv_info->devname);
+    if (strcmp (variable, "mode") == 0)
+    {
+	sprintf (sbuf, "%o", gv_info->info->mode);
+	return (sbuf);
+    }
+    if (strcmp (variable, "uid") == 0)
+    {
+	sprintf (sbuf, "%u", gv_info->info->uid);
+	return (sbuf);
+    }
+    if (strcmp (variable, "gid") == 0)
+    {
+	sprintf (sbuf, "%u", gv_info->info->gid);
+	return (sbuf);
+    }
     return (NULL);
 }   /*  End Function get_variable  */
-
